import re
import unittest

from pymdict.mongo_query_parser import MongoQueryParser


class MongoQueryParserTests(unittest.TestCase):

    def test_mongo_simple_query(self):
        m = MongoQueryParser()

        self.assertEqual(m.transform_request("val.hola = 33"), {'val.hola': 33.0})
        self.assertEqual(m.transform_request("val.hola eq 33"), {'val.hola': '33'})
        self.assertEqual(m.transform_request("val.hola > 33"), {'val.hola': {'$gt': 33.0}})
        self.assertEqual(m.transform_request("val.hola < 33"), {'val.hola': {'$lt': 33.0}})
        self.assertEqual(m.transform_request("val.hola >= 33"), {'val.hola': {'$gt': 32.0}})
        self.assertEqual(m.transform_request("val.hola <= 33"), {'val.hola': {'$lt': 34.0}})
        self.assertEqual(m.transform_request("val.hola != 33"), {'val.hola': {'$ne': 33.0}})
        self.assertEqual(m.transform_request("val.hola !eq 33"), {'val.hola': {'$ne': '33'}})
        self.assertEqual(m.transform_request("val.hola in ['33', '44']"), {'val.hola': {'$in': ['33', '44']}})
        self.assertEqual(m.transform_request("val.hola in [33, 44]"), {'val.hola': {'$in': [33, 44]}})
        self.assertEqual(m.transform_request("val.hola in ['hello jeje', '24 44']"), {'val.hola': {'$in': ['hello jeje', '24 44']}})
        self.assertEqual(m.transform_request("val.hola % thisregex"), {'val.hola': {'$regex': 'thisregex'}})
        self.assertEqual(m.transform_request("val.hola !% thisregex"), {'val.hola': {'$not': re.compile('thisregex')}})

    def test_mongo_complex_query(self):
        m = MongoQueryParser()

        self.assertEqual(m.transform_request("val.hola = 22 and val.pepe > 44"), {'$and': [{'val.hola': 22.0}, {'val.pepe': {'$gt': 44.0} }]})
        self.assertEqual(m.transform_request("val.hola = 22 and (val.pepe > 44 or val.juan < 44)"), {'$and': [{'val.hola': 22.0}, {'$or': [{'val.pepe': {'$gt': 44.0}}, {'val.juan': {'$lt': 44.0}}]}]})
        self.assertEqual(m.transform_request("(val.hola = 22 and val.pepe > 44) or val.juan < 44"), {'$or': [{'$and':[{'val.hola': 22.0}, {'val.pepe': {'$gt': 44.0}}]}, {'val.juan': {'$lt': 44.0}}]})


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2018 Iván de Paz Centeno
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from contextlib import contextmanager
from functools import partial

import pymongo
from bson import ObjectId
from pymongo import MongoClient, UpdateOne, DeleteOne
from pymongo.errors import ConnectionFailure, BulkWriteError
from pymdict.mongo_query_parser import MongoQueryParser


# Special collection for tracking information of mongo dictionaries, like forks families.
___MONGO_DICT_META___ = "___MONGO_DICT_META___"


class BasicMongoDict:
    """
    Mongo dictionary class.

    Allows to use a dictionary back-ended in a MongoDB.
    It allows to read, write, modify, create bulk operations and perform queries in a simplified way.

    [] Examples of classic-dict behaviour:

        * Create a new dictionary in MongoDB:
            >>> dictionary = MongoDict()

        * Insert elements in the dictionary:
            >>> dictionary['element'] = "value"

        * Read elements from the dictionary:
            >>> element = dictionary['element']

        * Iterate over keys:
            >>> for key in dictionary:
            ...     print(key)

        * Iterate over keys and values:
            >>> for key, value in dictionary.items():
            ...     print("{}: {}".format(key, value))

        * Remove element:
            >>> del dictionary[key]

      [] Examples of advanced features of MongoDict:

        * bulk write (write multiple elements in a single mongo call):
            >>> with dictionary.bulk(buffer_size=100) as d:
            ...     for x in range(10000):
            ...         d['key{}'.format(x)] = x

        * bulk remove:
            >>> with dictionary.bulk(buffer_size=100) as d:
            ...     for x in range(10000):
            ...         del d['key{}'.format(x)]

        * Perform a query:

            >>> for key, value, id in dictionary('key = key5'):
            ...     print("{}: {}".format(key, value))

        MongoDict allows to perform queries on the dictionary with a custom simple-to-use syntax instead of a mongo
        query.
        The basic syntax follows the next structure:

            Operand1 operator Operand2

        operator is defined in two ways:
            first level operators: "and", "or"
            second level operators: "=", "!=", ">", "<", ">=", "<=", "eq", "!eq", "%", "in"

        ----

        For example:

            >>> dictionary['a'] = 5
            >>> dictionary['b'] = 9
            >>> dictionary['d'] = 100
            >>>
            >>> query = "value > 5 and value < 100"
            >>> for k, v,_ in dictionary(query):
            ...     print("{}: {}".format(k, v))
            b: 9

            operators can be nested in a complex way by using brackets:
            >>> query = "value > 5 and (value < 100 or value = 100)"
            >>> for k, v,_ in dictionary(query):
            ...     print("{}: {}".format(k, v))
            b: 9
            d: 100

        For more information about queries, visit this wiki page: [TODO]
    """

    def __init__(self, original_dict_id:str=None, mongo_host:str="192.168.2.115", mongo_port:int=27017,
                 mongo_database="mongo_dicts", credentials:tuple=None):
        """
        Instantiates the dictionary back-ended in mongo
        :param original_dict_id: ID of the dictionary. This is going to be the name of the collection in the db.
                                 Reusing of an ID under a same mongo host, port and database leads to share the
                                 dictionary.
        :param mongo_host:  host of the mongodb
        :param mongo_port:  port of the mongodb
        :param mongo_database:  database name of the mongo db. By default it will use "mongo_dicts"
        :param credentials: tuple containing the user and password. Leave it as None if no credentials are required.
        """

        if original_dict_id is None:
            original_dict_id = str(ObjectId())

        mongo_uri = "mongodb://"

        if credentials is not None and len(credentials) == 2:
            mongo_uri += "{}:{}@".format(credentials[0], credentials[1])

        mongo_uri += "{}:{}".format(mongo_host, mongo_port)

        try:
            self._client = MongoClient(mongo_uri)
        except ConnectionFailure:
            raise ConnectionError("No connection to the remote dict backend. Ensure that a MongoDB backend is listening"
                                  " on {}:{}".format(mongo_host, mongo_port)) from None

        self._mongo_host = mongo_host
        self._mongo_port = mongo_port
        self._mongo_database = mongo_database
        self._credentials = credentials
        self._storage = self._client[mongo_database]
        self._instance = self._storage[original_dict_id]
        self._instance.create_index([('key', pymongo.TEXT)], name='key_index')

        self._original_dict_id = original_dict_id

    def __getitem__(self, item):
        """
        Retrieves the value for a given item.
        :param item: item key to search for.
        :return: value for the specified item key.
        """

        if type(item) is tuple:
            item = item[0]
            result = self._instance.find_one({'key': item})

            if result is None:
                raise KeyError(item)
        else:

            result = self._instance.find_one({'key': item}, {'_id': 0, 'value': 1})
            if result is None:
                raise KeyError(item)

            result = result['value']

        return result

    def __setitem__(self, key, value):
        """
        Sets the value for a given key.
        :param key: item key to set.
        :param value: value to set for the item.
        """
        self._on_modified_callback()
        self._instance.replace_one({'key': key},  {'key': key, 'value': value}, upsert=True)

    @contextmanager
    def bulk(self, buffer_size=100):
        """
        Performs a bulk operation over the dictionary. It can be write and/or delete of elements.
        It is managed by a contextmanager. An example of use is:
            >>> with dictionary.bulk() as d:
            ...     d['k1'] = "v1"
            ...     d['k2'] = "v2"
            ...     d['k3'] = "v3"

        :param buffer_size: size of the buffer to set for the bulk operation. This size is the threshold at which
                    the changes are commited to the backend. The reason is that bulk operations are limited in size, so
                    they must be buffered.
        :return: context manager for the write and delete bulk operations.
        """
        self._on_modified_callback()
        m = BulkMongoDict(self._original_dict_id, mongo_host=self._mongo_host, mongo_port=self._mongo_port,
                          buffer_size=buffer_size)
        yield m
        m.commit()

    def keys(self):
        """
        Retrieves a list of keys. Should not use this method if the dictionary is extremely big. Instead,
        iterate over it.
        :return: list of keys.
        """
        return list(self.__iter__())

    def __contains__(self, item):
        """
        Checks whether an item is contained in the instance.
        :param item: key value to check if it is contained or not.
        :return: boolean flag indicating if the key is contained or not.
        """
        return self._instance.find_one({'key': item}) != None

    def __str__(self):
        return "Mongo_Dict ({})".format(self._original_dict_id)

    def __repr__(self):
        return "Mongo_Dict ({})".format(self._original_dict_id)

    def values(self):
        """
        Returns the values of the dictionary. It must be forcefully iterated.
        """
        for x in self._instance.find({}, {'_id':0, 'value': 1}):
            yield x['value']

    def __iter__(self):
        for x in self._instance.find({}, {'_id':0, 'key': 1}):
            yield x['key']

    def __delitem__(self, key):
        self._on_modified_callback()
        result = self._instance.remove({'key': key})

    def items(self):
        for x in self._instance.find({}, {'_id':0, 'key': 1, 'value': 1}):
            yield x['key'], x['value']

    def __call__(self, query: str, count_only: bool=False):
        """
        Performs a query over the dictionary. It uses a simple query.
        :param query: string query to perform
        :param count_only: if set to true, it will return the length of the query result.
        :return: iterator for the elements that satisfies the query, or the size of the elements set if count_only
        param is true.
        """

        m = MongoQueryParser()
        mongo_query = m.transform_request(query)
        mongo_cursor = self._instance.find(mongo_query)

        if count_only:
            return mongo_cursor.count()
        else:
            for result in mongo_cursor:
                yield result['key'], result['value'], result['_id']

    def __len__(self):
        return self._instance.find().count()

    def __del__(self):
        self._client.fsync()
        self._client.close()

    def last_element_id(self):
        last_elements = list(self._instance.find().sort("_id", pymongo.DESCENDING).limit(1))

        result = last_elements[0]['_id'] if len(last_elements) > 0 else None
        return result

    def _on_modified_callback(self):
        pass

    def _drop(self):
        self._instance.drop()

    def get_my_id(self):
        return self._original_dict_id

    def update(self, ext_dict):
        with self.bulk() as b:
            for key, value in ext_dict.items():
                b[key] = value


class MongoDict(BasicMongoDict):
    """
    This dictionary is special: keeps track of itself in a special collection ___MONGO_DICT_META___
    """
    def __init__(self, original_dict_id:str=None, mongo_host:str="localhost", mongo_port:int=27017,
                 mongo_database="mongo_dicts", credentials:tuple=None, version=None, allow_morph:bool=True):
        BasicMongoDict.__init__(self, original_dict_id=original_dict_id, mongo_host=mongo_host, mongo_port=mongo_port,
                 mongo_database=mongo_database, credentials=credentials)

        self._allow_morph = allow_morph
        self._version = version
        self._load_version(version)

    def _on_modified_callback(self):
        dict_meta = BasicMongoDict(original_dict_id=___MONGO_DICT_META___, mongo_host=self._mongo_host,
                                   mongo_port=self._mongo_port,
                                   mongo_database=self._mongo_database, credentials=self._credentials)

        metadata = dict_meta[self._original_dict_id]
        metadata.update({'modified': True})
        dict_meta[self._original_dict_id] = metadata

    def _load_version(self, version=None):
        """
        Loads a specific version of this dict.
        :param version: version to load (int number from 0 to N). If None specified, it will load the latest version.
        """
        dict_meta = BasicMongoDict(original_dict_id=___MONGO_DICT_META___, mongo_host=self._mongo_host,
                                   mongo_port=self._mongo_port,
                                   mongo_database=self._mongo_database, credentials=self._credentials)

        try:
            metadata = dict_meta[self._original_dict_id]
        except KeyError:
            metadata = {'version': [], 'modified': True}
            dict_meta[self._original_dict_id] = metadata

        if version is None:
            # we pick last version
            if len(metadata['version']) == 0:
                self._version = 0
            else:
                self._version = metadata['version'][-1]

        else:
            self._version = version

        if self._version > 0:
            self._morph_into_fork(MongoDict(original_dict_id=self._original_dict_id, mongo_host=self._mongo_host,
                                            mongo_port=self._mongo_port, mongo_database=self._mongo_database,
                                            credentials=self._credentials, version=self._version-1))
        elif 'ancestor_fork' in metadata:
            self._morph_into_fork(MongoDict(original_dict_id=metadata['ancestor_fork'], mongo_host=self._mongo_host,
                                            mongo_port=self._mongo_port, mongo_database=self._mongo_database,
                                            credentials=self._credentials, version=metadata['ancestor_version']))


    """
    def _thread_checker(self):
        dict_meta = BasicMongoDict(original_dict_id=___MONGO_DICT_META___, mongo_host=self._mongo_host,
                                   mongo_port=self._mongo_port,
                                   mongo_database=self._mongo_database, credentials=self._credentials)

        while not self._exit:
            if dict_meta[self._original_dict_id]['version'][-1] != self._version:
                self._load_version()
            sleep(1)
    """

    def _morph_into_fork(self, father):
        """
        Morphs this dictionary into a fork dictionary class, if it is not already one.
        :param father: father of this fork.
        :return:
        """
        if self._allow_morph:

            if type(self) is not ForkedMongoDict:
                self.__class__ = ForkedMongoDict

                # Now we map relevant functions
                self.__contains__ = partial(ForkedMongoDict.__contains__, self)
                self.__str__ = partial(ForkedMongoDict.__str__, self)
                self.__repr__ = partial(ForkedMongoDict.__repr__, self)
                self.__len__ = partial(ForkedMongoDict.__len__, self)
                self.__getitem__ = partial(ForkedMongoDict.__getitem__, self)
                self.__delitem__ = partial(ForkedMongoDict.__delitem__, self)
                self.__call__ = partial(ForkedMongoDict.__call__, self)
                self.__iter__ = partial(ForkedMongoDict.__iter__, self)

                self.items = partial(ForkedMongoDict.items, self)
                self.bulk = partial(ForkedMongoDict.bulk, self)
                self._fork_father = father

            else:
                self._fork_father = father

        self._instance = self._storage[self._original_dict_id+("v{}".format(self._version) if self._version > 0 else "")]
        self._instance.create_index([('key', pymongo.TEXT)], name='key_index')

    def fork(self, new_id=None):
        if new_id is None:
            new_id = str(ObjectId())

        if new_id == self._original_dict_id:
            raise Exception("Fork cannot override father's ID")

        dict_meta = BasicMongoDict(original_dict_id=___MONGO_DICT_META___, mongo_host=self._mongo_host,
                                   mongo_port=self._mongo_port,
                                   mongo_database=self._mongo_database, credentials=self._credentials)

        metadata = dict_meta[self._original_dict_id]

        if metadata['modified']:
            metadata['version'].append(self._version + 1)
            metadata['modified'] = False
            dict_meta[self._original_dict_id] = metadata

        result = ForkedMongoDict(self, new_id, mongo_host=self._mongo_host, mongo_port=self._mongo_port,
                                 mongo_database=self._mongo_database, credentials=self._credentials)

        # Reload the latest version
        self._load_version()

        return result

    @contextmanager
    def bulk(self, buffer_size=100):
        """
        Performs a bulk operation over the dictionary. It can be write and/or delete of elements.
        It is managed by a contextmanager. An example of use is:
            >>> with dictionary.bulk() as d:
            ...     d['k1'] = "v1"
            ...     d['k2'] = "v2"
            ...     d['k3'] = "v3"

        :param buffer_size: size of the buffer to set for the bulk operation. This size is the threshold at which
                    the changes are commited to the backend. The reason is that bulk operations are limited in size, so
                    they must be buffered.
        :return: context manager for the write and delete bulk operations.
        """
        self._on_modified_callback()
        m = BulkMongoDict(self._original_dict_id, mongo_host=self._mongo_host, mongo_port=self._mongo_port,
                          buffer_size=buffer_size, version=self._version)
        yield m
        m.commit()


class ForkedMongoDict(MongoDict):
    """
    Dictionary that allows to work with an existing dict without overriding it.
    """

    def __init__(self, father:MongoDict, original_dict_id:str=None, mongo_host:str="localhost", mongo_port:int=27017,
                 mongo_database="mongo_dicts", credentials:tuple=None, version=None):
        MongoDict.__init__(self, original_dict_id=original_dict_id, mongo_host=mongo_host, mongo_port=mongo_port,
                           mongo_database=mongo_database, credentials=credentials, version=version)
        self._fork_father = MongoDict(father._original_dict_id, version=father._version, mongo_host=father._mongo_host,
                                      mongo_port=father._mongo_port, mongo_database=father._mongo_database,
                                      credentials=father._credentials)
        dict_meta = BasicMongoDict(original_dict_id=___MONGO_DICT_META___, mongo_host=self._mongo_host,
                                   mongo_port=self._mongo_port,
                                   mongo_database=self._mongo_database, credentials=self._credentials)

        metadata = dict_meta[self._original_dict_id]
        metadata['ancestor_fork'] = father._original_dict_id
        metadata['ancestor_version'] = father._version
        dict_meta[self._original_dict_id] = metadata

    def __contains__(self, item):
        """
        Checks whether an item is contained in the instance.
        :param item: key value to check if it is contained or not.
        :return: boolean flag indicating if the key is contained or not.
        """

        i = self._instance.find_one({'key': item})

        if i is None:
            return item in self._fork_father
        else:
            return "___removed" not in i

    def keys(self):
        current_keys = list(self._instance.find())
        removed_keys = set([k['key'] for k in current_keys if '___removed' in k])

        father_keys = self._fork_father.keys()
        total_keys = set(father_keys + [k['key'] for k in current_keys])
        return [r for r in total_keys if r not in removed_keys]

    def __str__(self):
        return "Mongo_Dict ({}) -- father: {} (v{})".format(self._original_dict_id, self._fork_father, self._fork_father._version)

    def __repr__(self):
        return "Mongo_Dict ({}) -- father: {} (v{})".format(self._original_dict_id, self._fork_father, self._fork_father._version)

    def __len__(self):
        new_elements = self._instance.find({'___removed': None}).count()
        removed_elements = self._instance.find({'___removed': 1}).count()

        return max(0, len(self._fork_father) - removed_elements) + new_elements

    def __getitem__(self, item):
        try:
            result = MongoDict.__getitem__(self, (item, ) if type(item) is not tuple else item)
        except KeyError:
            try:
                result = self._fork_father[(item, ) if type(item) is not tuple else item]
            except KeyError as e:
                raise e from None

        if result is None or (result is not None and '___removed' in result):
            raise KeyError(item)

        if type(item) is not tuple:
            result = result['value']

        return result

    def __delitem__(self, key):
        self._on_modified_callback()
        self._instance.replace_one({'key': key}, {'key': key, 'value': None, '___removed': 1}, upsert=True)

    @contextmanager
    def bulk(self, buffer_size=100):
        """
        Performs a bulk operation over the dictionary. It can be write and/or delete of elements.
        It is managed by a contextmanager. An example of use is:
            >>> with dictionary.bulk() as d:
            ...     d['k1'] = "v1"
            ...     d['k2'] = "v2"
            ...     d['k3'] = "v3"

        :param buffer_size: size of the buffer to set for the bulk operation. This size is the threshold at which
                    the changes are commited to the backend. The reason is that bulk operations are limited in size, so
                    they must be buffered.
        :return: context manager for the write and delete bulk operations.
        """
        m = BulkMongoDictForked(original_dict_id=self._original_dict_id, mongo_host=self._mongo_host,
                                mongo_port=self._mongo_port, mongo_database=self._mongo_database,
                                credentials=self._credentials, buffer_size=buffer_size, version=self._version)
        yield m
        self._on_modified_callback()
        m.commit()

    def items(self):
        for key, value, _id in self(""):
            yield key, value

    def __iter__(self):
        for key, _, _ in self(""):
            yield key

    def __call__(self, query:str, count_only:bool=False):
        """
        Performs a query over the dictionary. It uses a simple query.
        :param query: string query to perform
        :param count_only: if set to true, it will return the length of the query result.
        :return: iterator for the elements that satisfies the query, or the size of the elements set if count_only
        param is true.
        """

        RETRIEVE_BUFFER = 500
        # We perform a call this way we run through each element:

        m = MongoQueryParser()
        mongo_query = m.transform_request(query)
        mongo_query_for_removed = {'___removed': 1}
        mongo_query_for_edited_added = {'$and': [mongo_query, {'___removed': None}]}

        mongo_cursor_removed = self._instance.find(mongo_query_for_removed, {'key': 1, '_id': 0})
        mongo_cursor_edited = self._instance.find(mongo_query_for_edited_added)

        if count_only:
            return max(0, self._fork_father(query, count_only=True) - mongo_cursor_removed.count()) + mongo_cursor_edited.count()
        else:

            mongo_used = set()

            # First we yield the edited and added
            for result in mongo_cursor_edited:
                mongo_used.add(result['key'])
                yield result['key'], result['value'], result['_id']

                # Meanwhile we get RETRIEVE_BUFFER of removed elements
                for index, removed in zip(range(RETRIEVE_BUFFER), mongo_cursor_removed):
                    mongo_used.add(removed['key'])

            # Let's add the rest of the removed elements (if any yet) to the set
            for removed in mongo_cursor_removed:
                mongo_used.add(removed['key'])

            # Then we yield from the father's if it is not in the used list,
            mongo_cursor = self._fork_father(query)

            for key, value, _id in mongo_cursor:
                if key not in mongo_used:
                    yield key, value, _id


class BulkMongoDict(MongoDict):
    """
    Dictionary that allows bulk operations on a mongo dictionary.
    """
    def __init__(self, original_dict_id:str=None, mongo_host:str="localhost", mongo_port:int=27017,
                 mongo_database="mongo_dicts", credentials:tuple=None, buffer_size=100, version:int=None):
        self._buffer_size = buffer_size
        MongoDict.__init__(self, original_dict_id=original_dict_id, mongo_host=mongo_host,
                           mongo_port=mongo_port, mongo_database=mongo_database, credentials=credentials,
                           version=version, allow_morph=False)
        self._operations = []

    def __setitem__(self, key, value):
        self._on_modified_callback()
        self._operations.append(
            UpdateOne({"key": key}, {"$set": {"value": value}}, upsert=True)
        )

        if len(self._operations) > self._buffer_size:
            self.commit()

    def __delitem__(self, key):
        self._on_modified_callback()
        self._operations.append(
            DeleteOne({"key": key})
        )

        if len(self._operations) > self._buffer_size:
            self.commit()

    def commit(self):
        if len(self._operations) > 0:
            try:
                self._instance.bulk_write(self._operations, ordered=False)
                self._operations = []
            except BulkWriteError as ex:
                print("Could not write the bulk operations: {}".format(ex.details))
                raise


class BulkMongoDictForked(BulkMongoDict):

    def __delitem__(self, key):
        self._on_modified_callback()
        self._operations.append(
            UpdateOne({"key": key}, {"$set": {"value": None, "___removed": 1}}, upsert=True)
        )

        if len(self._operations) > self._buffer_size:
            self.commit()


class DictDropper:
    """
    Warning: this is a class capable of removing dictionaries! it might leave the dict-system in an inconsistent state!
    use it carefully!

    Why can it leave it in an inconsistent state?
    =============================================
    Dictionaries that fork from other dictionaries do not clone the content. Instead, they refer to the content of the
    original dictionary. If an original dictionary is removed and several forks are pointing to it, the forks will
    probably lose part of its content.
    """
    def __init__(self, mongo_host:str="localhost", mongo_port:int=27017,
                 mongo_database="mongo_dicts", credentials:tuple=None):
        self._mongo_host = mongo_host
        self._mongo_port = mongo_port
        self._mongo_database = mongo_database
        self._credentials = credentials

    def drop_dict(self, dict_id, remove_all_versions=True):
        dict_meta = BasicMongoDict(original_dict_id=___MONGO_DICT_META___, mongo_host=self._mongo_host,
                                   mongo_port=self._mongo_port,
                                   mongo_database=self._mongo_database, credentials=self._credentials)

        metadata = dict_meta[dict_id]

        versions = metadata['version']

        if remove_all_versions:
            for version in versions:
                BasicMongoDict(original_dict_id="{}v{}".format(dict_id, version), mongo_host=self._mongo_host,
                               mongo_port=self._mongo_port, mongo_database=self._mongo_database,
                               credentials=self._credentials)._drop()

            BasicMongoDict(original_dict_id=dict_id, mongo_host=self._mongo_host,
                           mongo_port=self._mongo_port,
                           mongo_database=self._mongo_database, credentials=self._credentials)._drop()
            del dict_meta[dict_id]
        else:
            MongoDict(original_dict_id=dict_id, mongo_host=self._mongo_host, mongo_port=self._mongo_port,
                      mongo_database=self._mongo_database, credentials=self._credentials)._drop()

            if len(metadata['version']) > 0:
                metadata['version'] = metadata['version'][:-1]
                dict_meta[dict_id] = metadata
            else:
                del dict_meta[dict_id]

        return len(versions) + 1



import unittest

from bson import ObjectId

from pymdict.mongo_dict import MongoDict, DictDropper


MONGO_HOST = "localhost"
MONGO_PORT = 27017


class CheckDict():
    
    def __init__(self, testcase_instance, dict_to_test):
        self._dict_to_test = dict_to_test
        self._testcase_instance = testcase_instance
        
    def test_len(self):
        d = self._dict_to_test

        self._testcase_instance.assertEqual(len(d), 0)
        d["1"] = 1
        self._testcase_instance.assertEqual(len(d), 1)

    def test_keys(self):
        d = self._dict_to_test

        self._testcase_instance.assertEqual(len(d), 0)
        d["1"] = 1
        self._testcase_instance.assertEqual(d.keys(), ["1"])

    def test_write_read(self):
        d = self._dict_to_test

        self._testcase_instance.assertEqual(len(d), 0)
        d["1"] = "my value"
        self._testcase_instance.assertEqual(d["1"], "my value")

        with self._testcase_instance.assertRaises(KeyError):
            d["5"]

    def test_write_read_complex_data(self):
        d = self._dict_to_test

        self._testcase_instance.assertEqual(len(d), 0)
        d["1"] = {"this": {"is": ["a very complex"], "piece": 33.4}, "of": b"data"}
        self._testcase_instance.assertEqual(d["1"], {"this": {"is": ["a very complex"], "piece": 33.4}, "of": b"data"})

    def test_contains(self):
        d = self._dict_to_test

        self._testcase_instance.assertEqual(len(d), 0)
        d["1"] = 22
        d["2"] = "foo"
        d["3"] = ["bar"]
        d[4] = ["bar"]

        self._testcase_instance.assertTrue("1" in d)
        self._testcase_instance.assertTrue("2" in d)
        self._testcase_instance.assertTrue("3" in d)
        self._testcase_instance.assertFalse("4" in d)
        self._testcase_instance.assertTrue(4 in d)

    def test_delete(self):
        d = self._dict_to_test

        self._testcase_instance.assertEqual(len(d), 0)
        d["1"] = 22
        d["2"] = "foo"
        d["3"] = ["bar"]

        self._testcase_instance.assertEqual(len(d), 3)

        self._testcase_instance.assertTrue("1" in d)
        self._testcase_instance.assertTrue("2" in d)

        del d["1"]

        self._testcase_instance.assertEqual(len(d), 2)

        self._testcase_instance.assertFalse("1" in d)
        self._testcase_instance.assertTrue("2" in d)

    def test_iter(self):
        d = self._dict_to_test

        d["1"] = 22
        d["2"] = "foo"
        d["3"] = ["bar"]

        keys = {"1", "2", "3"}

        for x in d:
            self._testcase_instance.assertIn(x, keys)

    def test_items(self):
        d = self._dict_to_test

        d["1"] = 22
        d["2"] = "foo"
        d["3"] = ["bar"]

        normal_dict = {
            "1": 22,
            "2": "foo",
            "3": ["bar"]
        }

        for key, value in d.items():
            self._testcase_instance.assertEqual(value, normal_dict[key])

    def test_update(self):
        d = self._dict_to_test

        self._testcase_instance.assertEqual(len(d), 0)

        normal_dict = {
            "1": 22,
            "2": "foo",
            "3": ["bar"],
            "4": 44,
            "5": 55
        }

        d.update(normal_dict)

        self._testcase_instance.assertEqual(len(d), 5)

        for key, value in normal_dict.items():
            self._testcase_instance.assertEqual(d[key], value)

    def test_query(self):
        d = self._dict_to_test

        normal_dict = {
            "1": 22,
            "2": "foo",
            "3": ["bar"]
        }

        for key, value, _id in d(""):
            self._testcase_instance.assertEqual(value, normal_dict[key])
            self._testcase_instance.assertEqual(type(_id), ObjectId)

        for key, value, _id in d("value = 22"):
            self._testcase_instance.assertEqual(key, "1")
            self._testcase_instance.assertEqual(value, normal_dict[key])

        for key, value, _id in d("value = 22"):
            self._testcase_instance.assertEqual(key, "1")
            self._testcase_instance.assertEqual(value, normal_dict[key])


class MongoDictTests(unittest.TestCase):

    def setUp(self):
        self.m = MongoDict(mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)

    def test_mongo_dict_len(self):
        test = CheckDict(self, self.m)
        test.test_len()

    def test_mongo_dict_keys(self):
        test = CheckDict(self, self.m)
        test.test_keys()

    def test_mongo_dict_write_read(self):
        test = CheckDict(self, self.m)
        test.test_write_read()

    def test_mongo_dict_write_read_complex_data(self):
        test = CheckDict(self, self.m)
        test.test_write_read_complex_data()

    def test_mongo_dict_contains(self):
        test = CheckDict(self, self.m)
        test.test_contains()

    def test_mongo_dict_delete(self):
        test = CheckDict(self, self.m)
        test.test_delete()

    def test_iter(self):
        test = CheckDict(self, self.m)
        test.test_iter()

    def test_items(self):
        test = CheckDict(self, self.m)
        test.test_items()

    def test_update(self):
        test = CheckDict(self, self.m)
        test.test_update()

    def test_query(self):
        test = CheckDict(self, self.m)
        test.test_query()

    def tearDown(self):
        self.dropper = DictDropper(mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)
        self.dropper.drop_dict(self.m.get_my_id())


class ForkedMongoDictTests(unittest.TestCase):

    def setUp(self):
        self.original = MongoDict(mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)
        self.fork = self.original.fork()

    def test_mongo_dict_len(self):
        test = CheckDict(self, self.fork)
        test.test_len()

    def test_mongo_dict_keys(self):
        test = CheckDict(self, self.fork)
        test.test_keys()

    def test_mongo_dict_write_read(self):
        test = CheckDict(self, self.fork)
        test.test_write_read()

    def test_mongo_dict_write_read_complex_data(self):
        test = CheckDict(self, self.fork)
        test.test_write_read_complex_data()

    def test_mongo_dict_contains(self):
        test = CheckDict(self, self.fork)
        test.test_contains()

    def test_mongo_dict_delete(self):
        test = CheckDict(self, self.fork)
        test.test_delete()

    def test_iter(self):
        test = CheckDict(self, self.fork)
        test.test_iter()

    def test_items(self):
        test = CheckDict(self, self.fork)
        test.test_items()

    def test_update(self):
        test = CheckDict(self, self.fork)
        test.test_update()

    def test_query(self):
        test = CheckDict(self, self.fork)
        test.test_query()

    def tearDown(self):
        self.dropper = DictDropper(mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)
        self.dropper.drop_dict(self.original.get_my_id())
        self.dropper.drop_dict(self.fork.get_my_id())

class ForkedMongoDictTests2(unittest.TestCase):

    def setUp(self):
        self.original = MongoDict(mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)
        self.original["val1"] = 55
        self.original["val2"] = "hello"
        self.fork = self.original.fork()
        del self.fork["val1"]
        del self.fork["val2"]

    def _assert_original_kept(self):
        self.assertEqual(len(self.original), 2)
        self.assertEqual(self.original['val1'], 55)
        self.assertEqual(self.original['val2'], "hello")

    def test_mongo_dict_len(self):
        test = CheckDict(self, self.fork)
        test.test_len()
        self._assert_original_kept()

    def test_mongo_dict_keys(self):
        test = CheckDict(self, self.fork)
        test.test_keys()
        self._assert_original_kept()

    def test_mongo_dict_write_read(self):
        test = CheckDict(self, self.fork)
        test.test_write_read()
        self._assert_original_kept()

    def test_mongo_dict_write_read_complex_data(self):
        test = CheckDict(self, self.fork)
        test.test_write_read_complex_data()
        self._assert_original_kept()

    def test_mongo_dict_contains(self):
        test = CheckDict(self, self.fork)
        test.test_contains()
        self._assert_original_kept()

    def test_mongo_dict_delete(self):
        test = CheckDict(self, self.fork)
        test.test_delete()
        self._assert_original_kept()

    def test_iter(self):
        test = CheckDict(self, self.fork)
        test.test_iter()
        self._assert_original_kept()

    def test_items(self):
        test = CheckDict(self, self.fork)
        test.test_items()
        self._assert_original_kept()

    def test_update(self):
        test = CheckDict(self, self.fork)
        test.test_update()
        self._assert_original_kept()

    def test_query(self):
        test = CheckDict(self, self.fork)
        test.test_query()
        self._assert_original_kept()

    def tearDown(self):
        self.dropper = DictDropper(mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)
        self.dropper.drop_dict(self.original.get_my_id())
        self.dropper.drop_dict(self.fork.get_my_id())


class ForkedMongoDictTests3(unittest.TestCase):

    def setUp(self):
        self._drop_db()
        self.original = MongoDict("original", mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)
        self.original["val1"] = 55
        self.original["val2"] = "hello"
        self.fork1 = self.original.fork("fork1")
        self.fork1["val3"] = 65
        self.fork2 = self.fork1.fork("fork2")
        del self.fork2["val1"]
        del self.fork2["val2"]
        del self.fork2["val3"]

    def _assert_original_kept(self):
        self.assertEqual(len(self.original), 2)
        self.assertEqual(self.original['val1'], 55)
        self.assertEqual(self.original['val2'], "hello")

        self.assertEqual(len(self.fork1), 3)
        self.assertEqual(self.fork1['val1'], 55)
        self.assertEqual(self.fork1['val2'], "hello")
        self.assertEqual(self.fork1['val3'], 65)

    def test_mongo_dict_len(self):
        test = CheckDict(self, self.fork2)
        test.test_len()
        self._assert_original_kept()

    def test_mongo_dict_keys(self):
        test = CheckDict(self, self.fork2)
        test.test_keys()
        self._assert_original_kept()

    def test_mongo_dict_write_read(self):
        test = CheckDict(self, self.fork2)
        test.test_write_read()
        self._assert_original_kept()

    def test_mongo_dict_write_read_complex_data(self):
        test = CheckDict(self, self.fork2)
        test.test_write_read_complex_data()
        self._assert_original_kept()

    def test_mongo_dict_contains(self):
        test = CheckDict(self, self.fork2)
        test.test_contains()
        self._assert_original_kept()

    def test_mongo_dict_delete(self):
        test = CheckDict(self, self.fork2)
        test.test_delete()
        self._assert_original_kept()

    def test_iter(self):
        test = CheckDict(self, self.fork2)
        test.test_iter()
        self._assert_original_kept()

    def test_items(self):
        test = CheckDict(self, self.fork2)
        test.test_items()
        self._assert_original_kept()

    def test_update(self):
        test = CheckDict(self, self.fork2)
        test.test_update()
        self._assert_original_kept()

    def test_query(self):
        test = CheckDict(self, self.fork2)
        test.test_query()
        self._assert_original_kept()

    def _drop_db(self):
        try:
            self.dropper = DictDropper(mongo_host=MONGO_HOST, mongo_port=MONGO_PORT)
            self.dropper.drop_dict("original")
            self.dropper.drop_dict("fork1")
            self.dropper.drop_dict("fork2")

        except:
            pass

    def tearDown(self):
        self._drop_db()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2018 Iván de Paz Centeno
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re


class MongoQueryParser():
    """
    Allows to process a string query and to convert it into a mongo query.
    """

    def __init__(self):
        pass


    def transform_request(self, str_query:str):
        """
        Transforms the given request in a mongo query

        :param str_query: String containing the query.

        A request is a set of conditions that elements must satisfy. For example, if we have the following dictionary:

        {
            "father": { "name": "foo", "age": 44},
            "mother": { "name": "bar", "age": 41},
            "sister": { "name": "foobar", "age": 19}
        }

        and it is wanted to get the element with name "foo", it can be done as follows:

            value.name eq "foo"

        This query will return the "father" element.

        Queries can be combined in such a way to make complex queries. Example:

            value.name eq "foo" or (value.age > 40 and value.age < 50)

        This query returns the "father" and "mother" elements.

        Available operators are:

            eq               equals for comparing text
            =                equals for comparing numbers
            !                negation, can be prepended to eq and =, for example: !eq and !=
            >                greater than a number
            <                lesser than a number
            >=               greater or equal than a number
            <=               lesser or equal than a number
            %                Regex match
            !%               Regex not match
            in               in a list of values

            or               To join two conditions in OR
            and              To join two conditions in AND

        :return: JSON query for a MongoDB
        """
        if str_query == "":
            return {}

        splits = self._do_split(str_query, open_splits=["(", "[", "'"], close_splits=[")", "]", "'"],
                                special_separators="[")
        ops = self._retrieve_ops_tree(splits, open_splits=["(", "[", "'"], close_splits=[")", "[", "'"])
        mongo = self._to_mongo_query_dict(ops)

        return mongo

    @staticmethod
    def _do_split(text:str, separator:str=" ", open_splits:list=None, close_splits:list=None, special_separators:list=None):
        """
        Splits a text into chunks. Same functionality as string.split() but it allows to split text in a more complex
        way.


        :param text: text to split
        :param separator: separator to use. For example, a " " will split the text in chunks separated by spaces.
        :param open_splits:  If a more complex separation is required, for example those between brackets, then here
        it can be specified the open token
        :param close_splits:
        :param special_separators:
        :return:
        """

        if open_splits is None:
            open_splits = []

        if close_splits is None:
            close_splits = []

        if special_separators is None:
            special_separators = []

        segments = []
        segment = ""

        l_2 = ""
        index = 0

        while index < len(text):
            l_1 = l_2
            l_2 = text[index]

            if l_2 in open_splits and (l_1 == '' or l_1 == separator or separator == ""):
                open_split_index = open_splits.index(l_2)

                segment = MongoQueryParser._do_encapsulated_split(text[index:], sep_init=l_2, sep_end=close_splits[open_split_index], include_separators=l_2 in special_separators)
                segments.append(segment)
                index += len(segment) + 2
                segment = ""
            elif l_2 == separator:
                segments.append(segment)
                segment = ""
            else:
                segment += l_2

            index += 1

        if len(segment) > 0:
            segments.append(segment)

        return segments

    @staticmethod
    def _do_encapsulated_split(text:str, sep_init="(", sep_end=")", include_separators=False):
        """
        Splits by text between tokens.
        :param text: text to split
        :param sep_init: token that specifies the start of the segment
        :param sep_end: token that specifies the end of the segment
        :param include_separators: boolean flag that specifies if the token separators should be included in the
        segment or not.
        :return: split segments
        """
        segment = ""
        open = False
        ESCAPE_CHAR = '\\'
        previous = ""
        for index, l in enumerate(text):

            if l == sep_init and not open and previous != ESCAPE_CHAR:
                open = True

            elif l == sep_end and previous != ESCAPE_CHAR:
                break

            elif open:
                segment += l
            previous = l

        return segment if not include_separators else sep_init + segment + sep_end


    def _retrieve_ops2_tree(splits:list):
        """
        Retrieves the ops tree for second level operations.
        :return:
        """
        first_level_operators = [">", "<", "=", "!=", ">=", "<=", "eq", "!eq", "in", "%", "!%"]

        # splits should be a 3 elements list : operand1 operator operand2
        if len(splits) < 3:
            raise Exception("Invalid syntax on {}".format(" ".join(splits)))

        operator = splits[1]
        if operator not in first_level_operators:
            raise Exception("Invalid operator on {}".format(" ".join(splits)))

        previous_oper = splits[0:1]
        next_oper = splits[2:]
        return [operator, previous_oper, next_oper]

    @staticmethod
    def _retrieve_ops_tree(splits:list, open_splits:list=None, close_splits:list=None):
        """
        Retrieves the ops tree from a split query.
        :param splits: split segments from the query.
        :param open_splits: token that specifies the open of the splits, forwarded to the 'do_split' function.
        :param close_splits: token that specifies the close of the splits, forwarded to the 'do_split' function.
        :return:
        """
        second_level_operators = ["and", "or"]

        operators = {operator: [] for operator in second_level_operators+ [None]}

        previous_index = 0
        last_operand = None

        for index, operand in enumerate(splits):
            if operand in operators:
                last_operand = operand
                previous_oper = splits[previous_index:index]

                if len(previous_oper) > 1:
                    previous_oper = MongoQueryParser._retrieve_ops2_tree(previous_oper)

                elif len(previous_oper) == 1 and previous_oper[0][0] != "[":
                        previous_oper = MongoQueryParser._retrieve_ops_tree(MongoQueryParser._do_split(previous_oper[0].strip(), open_splits=open_splits, close_splits=close_splits))

                operators[operand].append(previous_oper)
                previous_index = index + 1

        if previous_index < index + 1:
            previous_oper = splits[previous_index:]

            if len(previous_oper) > 1:
                previous_oper = MongoQueryParser._retrieve_ops2_tree(previous_oper)
            else:
                previous_oper = MongoQueryParser._retrieve_ops_tree(MongoQueryParser._do_split(previous_oper[0], open_splits=open_splits, close_splits=close_splits))

            operators[last_operand].append(previous_oper)

        return operators

    @staticmethod
    def _to_mongo_query_list(ops_list:list):
        """
        Converts a list of operations into a Mongo query.
        This method is recursively called by the method _to_mongo_query_dict() if it finds any

        :param ops_list: list of mongo operations to convert.
        :return: Mongo query.
        """

        [operation, operand1, operand2] = ops_list
        operand1 = operand1[0]
        operand2 = operand2[0]

        result = {}

        if operation == "eq":
            result[operand1] = operand2
        elif operation == "!eq":
            result[operand1] = {"$ne": operand2}
        elif operation == "=":
            result[operand1] = float(operand2)
        elif operation == "!=":
            result[operand1] = {"$ne": float(operand2)}
        elif operation == ">":
            result[operand1] = {"$gt": float(operand2)}
        elif operation == "<":
            result[operand1] = {"$lt": float(operand2)}
        elif operation == "<=":
            result[operand1] = {"$lt": float(operand2)+1}
        elif operation == ">=":
            result[operand1] = {"$gt": float(operand2)-1}
        elif operation == "%":
            result[operand1] = {"$regex": operand2}
        elif operation == "!%":
            result[operand1] = {"$not": re.compile(operand2)}
        elif operation == "in":
            splits = MongoQueryParser._do_split(operand2[1:-1], separator="", open_splits=["'"], close_splits=["'"])
            if len(splits) == 1:
                try:
                    splits = [float(x) for x in MongoQueryParser._do_split(operand2[1:-1], separator=",")]
                except:
                    pass

            result[operand1] = {"$in": splits}

        return result

    @staticmethod
    def _to_mongo_query_dict(ops_dict):
        """
        Converts an ops tree into a Mongo query.
        :param ops_dict: Ops tree, extracted from the _retrieve_ops_tree()
        :return: Mongo query.
        """

        ors = ops_dict['or']
        ands = ops_dict['and']
        base = ops_dict[None]

        result = {}
        if len(ors) > 0:
            list_ors = []

            for op in ors:
                if type(op) is list:
                    l = MongoQueryParser._to_mongo_query_list(op)
                elif type(op) is dict:
                    l = MongoQueryParser._to_mongo_query_dict(op)

                list_ors += [l]

            result['$or'] = list_ors

        if len(ands) > 0:
            list_ands = []
            for op in ands:
                if type(op) is list:
                    l = MongoQueryParser._to_mongo_query_list(op)
                elif type(op) is dict:
                    l = MongoQueryParser._to_mongo_query_dict(op)
                list_ands += [l]

            if len(result) > 0:
                result['$or'] += list_ands
            else:
                result['$and'] = list_ands

        if len(base) > 0:
            result = MongoQueryParser._to_mongo_query_list(base[0])

        return result






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# MIT License
#
# Copyright (c) 2018 Iván de Paz Centeno
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
from setuptools import setup, setuptools
from pymdict import __version__


__author__ = 'Iván de Paz Centeno'


def readme():
    with open('README.rst', encoding="UTF-8") as f:
        return f.read()


if sys.version_info < (3, 4, 1):
    sys.exit('Python < 3.4.1 is not supported!')


setup(name='pymdict',
      version=__version__,
      description='Python Mongo-Dict, a dictionary-like class automatically stored in mongo, with query capabilities.',
      long_description=readme(),
      url='http://github.com/ipazc/pymdict',
      author='Iván de Paz Centeno',
      author_email='ipazc@unileon.es',
      license='MIT',
      packages=setuptools.find_packages(),
      install_requires=[
          "pymongo"
      ],
      classifiers=[
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ],
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
      keywords="pymdict mongo dict python query advanced fork pip package",
      zip_safe=False)
